// Fichier Registry.cpp

#include "registry.h"
#include "ConversionChaines.h"
#include <string>
using std::string;
using std::wstring;

#include <objbase.h>
#include <cassert>

struct REG_DATA
{
   const wchar_t *pszKey;
   const wchar_t *pszValue;
   const wchar_t *pszData;
};

// Dfinit la cl donne et sa valeur.
bool FixerCleEtValeur
   (const wstring & Cle, const wstring & SousCle, const wstring & Valeur);

// Ouvre une cl et dfinit une valeur.
bool FixerValeurDansCle
   (const wstring & Cle, const wstring & ValeurNommee, const wstring &Valeur);

// Convertit un CLSID en une chaine de caractres.
wstring CLSIDtochar(REFCLSID clsid);

// Supprime szKeyChild et tous ses descendants.
LONG recursiveDeleteKey(HKEY hKeyParent, const wstring & CleEnfant);

HRESULT UnregisterServerEx(const REG_DATA regData[]);


std::wstring ObtenirNomCompletModule (HMODULE hModule)
{
   // Taille d'un CLSID en tant que chaine.
   const int TAILLE_MAX_NOMFICHIER = MAX_PATH + 1; // MAX_PATH est dfini par Windows
   wchar_t NomModule[TAILLE_MAX_NOMFICHIER] = { 0 };
   DWORD dwResult = GetModuleFileName(hModule, NomModule, sizeof(NomModule)/sizeof(NomModule[0]));
   assert(dwResult);
   return NomModule;
}

// Convertit le CLSID en un char.
wstring ContruireCLSIDBrut (REFCLSID clsid)
   { return CLSIDtochar(clsid); }

wstring ContruireCLSID (const wstring & clsid_brut)
{
   // construit la cl CLSID\\{...}
   static const wstring PREFIXE_CLSID = L"CLSID\\";
   return PREFIXE_CLSID + clsid_brut;
}

bool EnregistrementCle (const wstring & Cle, const wstring & NomConvivial)
   { return FixerCleEtValeur (Cle, wstring (), NomConvivial); }

bool EnregistrementProgID
   (const wstring & Cle, const wstring & ProgID)
{
   static const wstring SOUSCLE_PROGID = L"ProgID";
   return FixerCleEtValeur (Cle, SOUSCLE_PROGID, ProgID);
}

bool EnregistrementTypeServeur
   (const wstring & NomModule, const wstring & NomModuleComplet,
    const wstring & Cle, const wstring & ThreadingModel)
{
   static const wstring EXTENSION_SERVEUR_EXTERNE = L".exe";
   static const wstring PREFIXE_THREADING_MODEL = L"ThreadingModel";
   static const wstring SEPARATEUR = L"\\";
   // Ajoute la sous-cl "nom de fichier du serveur" sous la cl CLSID.
   bool ResultatEnregistrement;
   if(NomModule.find (EXTENSION_SERVEUR_EXTERNE) == wstring::npos)
   {
      static const wstring NOMCLE_SERVEUR_INTERNE = L"InprocServer32";
      wstring CleInterne = Cle + SEPARATEUR + NOMCLE_SERVEUR_INTERNE;
      ResultatEnregistrement =
         FixerCleEtValeur   (Cle, NOMCLE_SERVEUR_INTERNE, NomModuleComplet) &&
         FixerValeurDansCle (CleInterne, PREFIXE_THREADING_MODEL, ThreadingModel);
   }
   else
   {
      static const wstring NOMCLE_SERVEUR_EXTERNE = L"LocalServer32";
      ResultatEnregistrement =
         FixerCleEtValeur (Cle, NOMCLE_SERVEUR_EXTERNE, NomModuleComplet);
   }
   return ResultatEnregistrement;
}

bool EnregistrementVersionDependent
   (const wstring & ProgID, const wstring & NomConvivial, const wstring & clsid_brut)
{
   static const wstring PREFIXE_CLSID = L"CLSID";
   return FixerCleEtValeur (ProgID, wstring(), NomConvivial) &&
          FixerCleEtValeur (ProgID, PREFIXE_CLSID, clsid_brut);
}

bool EnregistrementCleVersionIndependent
   (const wstring & Cle, const wstring & NomIndependantDeVersion)
{
   static const wstring SOUSCLE_VERSIONINDEPENDENTPROGID = L"VersionIndependentProgID";
   return FixerCleEtValeur (Cle, SOUSCLE_VERSIONINDEPENDENTPROGID,
                            NomIndependantDeVersion);
}

bool EnregistrementVersionIndependent
   (const wstring & ProgID,     const wstring & NomConvivial,
    const wstring & clsid_brut, const wstring & NomIndependantDeVersion)
{
   static const wstring PREFIXE_CLSID = L"CLSID";
   static const wstring PREFIXE_CURRENT_VERSION = L"CurVer";
   return FixerCleEtValeur (NomIndependantDeVersion, wstring(), NomConvivial) && 
          FixerCleEtValeur (NomIndependantDeVersion, PREFIXE_CLSID, clsid_brut) &&
          FixerCleEtValeur (NomIndependantDeVersion, PREFIXE_CURRENT_VERSION, ProgID);
}

// Inscrit le composant dans le registre
HRESULT RegisterServer
   (const string & _NomModule, // nom du module, qu'il soit DLL ou EXE
    REFCLSID clsid,               // ID de classe
    const string & _NomConvivial,   // nom convivial
    const string & _NomIndependantDeVersion,   // Programmatique
    const string & _ProgID,         // ID
    const char * _ThreadingModel) // modle multithread ou non
{
   // obtient l'emplacement du serveur
   const wstring NomModule (_NomModule.begin (), _NomModule.end ());
   const wstring NomConvivial (_NomConvivial.begin (), _NomConvivial.end ());
   const wstring NomIndependantDeVersion (_NomIndependantDeVersion.begin (), _NomIndependantDeVersion.end ());
   const wstring ProgID (_ProgID.begin (), _ProgID.end ());
   const wstring ThreadingModel = StringToWString (_ThreadingModel? _ThreadingModel : "");

   HMODULE hMod = GetModuleHandle(NomModule.c_str ());
   if (!hMod)
   {
      DWORD err = ::GetLastError (); // debugging
   }
   const wstring NomModuleComplet = ObtenirNomCompletModule (hMod);

   const wstring clsid_brut = ContruireCLSIDBrut (clsid);
   const wstring Cle = ContruireCLSID (clsid_brut);

   // Ajoute le CLSID au registre.
   HRESULT hr = S_OK;
   if (!EnregistrementCle (Cle, NomConvivial) ||
       !EnregistrementTypeServeur (NomModule, NomModuleComplet, Cle, ThreadingModel) ||
       !EnregistrementProgID (Cle, ProgID) ||
       !EnregistrementCleVersionIndependent (Cle, NomIndependantDeVersion) ||
       !EnregistrementVersionIndependent    (ProgID, NomConvivial, clsid_brut, NomIndependantDeVersion) ||
       !EnregistrementVersionDependent      (ProgID, NomConvivial, clsid_brut))
      hr = S_FALSE;

   return hr;
}

// Supprime le composant du registre.
LONG UnregisterServer
   (REFCLSID clsid,             // ID de classe
    const string & _NomIndependantDeVersion, // Programmatique
    const string & _ProgID)       // ID
{
   const wstring NomIndependantDeVersion
      (_NomIndependantDeVersion.begin (), _NomIndependantDeVersion.end ());
   const wstring ProgID (_ProgID.begin (), _ProgID.end ());

   const wstring clsid_brut = ContruireCLSIDBrut (clsid);
   const wstring Cle = ContruireCLSID (clsid_brut);

   // supprime la cl CLSID\{...}
   LONG lResult = recursiveDeleteKey(HKEY_CLASSES_ROOT, Cle);
   assert((lResult == ERROR_SUCCESS) || (lResult == ERROR_FILE_NOT_FOUND)); // Subkey may not exist.

   // supprime la cl version-independent ProgID.
   lResult = recursiveDeleteKey(HKEY_CLASSES_ROOT, NomIndependantDeVersion);
   assert((lResult == ERROR_SUCCESS) || (lResult == ERROR_FILE_NOT_FOUND)); // Subkey may not exist.

   // supprime la cl ProgID.
   lResult = recursiveDeleteKey(HKEY_CLASSES_ROOT, ProgID);
   assert((lResult == ERROR_SUCCESS) || (lResult == ERROR_FILE_NOT_FOUND)); // Subkey may not exist.

   return S_OK;
}

// Convertit un CLSID en une chaine.
std::wstring CLSIDtochar(REFCLSID clsid)
{
   // Get CLSID
   LPOLESTR wszCLSID = 0;
   HRESULT hr = StringFromCLSID(clsid, &wszCLSID);
   assert(SUCCEEDED(hr));
   
   wstring clsid_obtenu = wszCLSID;

   // Libre la mmoire.
   CoTaskMemFree(wszCLSID);

   return clsid_obtenu;
}

// supprime une cl et tous ses descendants.
LONG recursiveDeleteKey
   (HKEY hKeyParent,           // Parent de la cl  supprimer
    const wstring & CleEnfant)  // cl  supprimer
{
   // Ouvre l'enfant.
   HKEY hKeyEnfant;
   LONG lRes = RegOpenKeyEx(hKeyParent, CleEnfant.c_str (), 0, KEY_ALL_ACCESS, &hKeyEnfant);
   if(lRes != ERROR_SUCCESS)
      return lRes;

   // Enumre tous les decendents de cet enfant.
   FILETIME time;
   const int TAILLE_TAMPON = 256;
   wchar_t szBuffer[TAILLE_TAMPON] = { 0 };
   DWORD dwSize = TAILLE_TAMPON;
   while(RegEnumKeyEx(hKeyEnfant, 0, szBuffer, &dwSize, 0, 0, 0, &time) == S_OK)
   {
      // supprime les decendants de cet enfant.
      lRes = recursiveDeleteKey(hKeyEnfant, szBuffer);
      if(lRes != ERROR_SUCCESS)
      {
         // nettoyage avant fin.
         RegCloseKey(hKeyEnfant);
         return lRes;
      }
      dwSize = TAILLE_TAMPON;
   }

   // Ferme l'enfant.
   RegCloseKey(hKeyEnfant);

   // Supprime cet enfant.
   return RegDeleteKey(hKeyParent, CleEnfant.c_str ());
}

// Cre une cl et dfinit sa valeur.
bool FixerCleEtValeur
   (const wstring & Cle, const wstring &SousCle, const wstring &Valeur)
{
   wstring KeyBuf = Cle;

   // Ajoute nom de sous-cl au tampon.
   if(!SousCle.empty())
   {
      static const wstring PREFIXE = L"\\";
      KeyBuf += PREFIXE;
      KeyBuf += SousCle;
   }

   // Cre et ouvre cl plus sous-cl.
   HKEY hKey;
   long lResult = RegCreateKeyEx(HKEY_CLASSES_ROOT, KeyBuf.c_str (), 0, 0, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey, 0);
   if(lResult != ERROR_SUCCESS)
      return false;

   // dfinit la valeur.
   if(!Valeur.empty ())
      RegSetValueEx (hKey, NULL, 0, REG_SZ,
                     reinterpret_cast<const BYTE *>(Valeur.c_str ()),
                     static_cast<DWORD>(Valeur.length () + 1) * sizeof(std::wstring::value_type));

   RegCloseKey(hKey);
   return true;
}

// Ouvre une cl et dfinit une valeur.
bool FixerValeurDansCle
   (const wstring & Cle, const wstring & ValeurNommee, const wstring &Valeur)
{
   wstring KeyBuf = Cle;

   // Cre et ouvre cl plus sous-cl.
   HKEY hKey;
   long lResult = RegOpenKeyEx(HKEY_CLASSES_ROOT, KeyBuf.c_str (), 0, KEY_SET_VALUE, &hKey);
   if(lResult != ERROR_SUCCESS)
      return false;

    // dfinit la valeur.
   if(!Valeur.empty())
      RegSetValueEx (hKey, ValeurNommee.c_str (), 0, REG_SZ,
                     reinterpret_cast<const BYTE*>(Valeur.c_str ()),
                     static_cast<DWORD>(Valeur.length () + 1) * sizeof(std::wstring::value_type));

   RegCloseKey(hKey);
   return true;
}

HRESULT RegisterServerEx(const REG_DATA regData[], const wchar_t* szModuleName)
{
   int Cpt = 0;
   while (regData[Cpt].pszKey || regData[Cpt].pszValue || regData[Cpt].pszData)
   {
      const wchar_t *pszKey = regData[Cpt].pszKey;
      const wchar_t *pszValue = regData[Cpt].pszValue;
      const wchar_t *pszData = regData[Cpt].pszData;

      if(pszData == reinterpret_cast<const wchar_t*>(-1))
         pszData = szModuleName;

      HKEY hKey;
      long err = RegCreateKey(HKEY_CLASSES_ROOT, pszKey, &hKey);
      if(err == ERROR_SUCCESS)
      {
         err = RegSetValueEx(hKey, pszValue, 0, REG_SZ, reinterpret_cast<const BYTE*>(pszData),
                             static_cast<DWORD>(std::wstring (pszData).length () + 1));
         RegCloseKey(hKey);
      }
      if(err != ERROR_SUCCESS)
      {
         UnregisterServerEx(regData);
         return REGDB_E_WRITEREGDB;
      }
      ++Cpt;
   }
   return S_OK;
}

HRESULT UnregisterServerEx(const REG_DATA regData[])
{
   HRESULT hr = S_OK;
   int nEntries = 0;
   while (regData[nEntries].pszKey || regData[nEntries].pszValue || regData[nEntries].pszData)
      ++nEntries;

   for(int count = nEntries - 1; count >= 0; count--)
      if(RegDeleteKey(HKEY_CLASSES_ROOT, regData[count].pszKey) != ERROR_SUCCESS)
         hr = S_FALSE;

   return hr;
}